//
//  tLBulletSoftBodyPlugIn.m
//  tLBulletSoftBody
//
//  Created by pascal on 5/22/11.
//  Copyright (c) 2011 t-linked.com . All rights reserved.
//

/* It's highly recommended to use CGL macros instead of changing the current context for plug-ins that perform OpenGL rendering */
#import <OpenGL/gluMacro.h>
#import "tLBulletSoftBodyPlugIn.h"

#include "TLUtils.h"

#include "btBulletDynamicsCommon.h"

#include "LinearMath/btAlignedObjectArray.h"
#include "LinearMath/btConvexHull.h"
#include "LinearMath/btQuickprof.h"
#include "LinearMath/btIDebugDraw.h"

#include "BulletSoftBody/btSoftBody.h"
#include "BulletSoftBody/btSoftRigidDynamicsWorld.h"
#include "BulletSoftBody/btSoftBodyRigidBodyCollisionConfiguration.h"
#include "BulletSoftBody/btSoftBodyHelpers.h"

#include "BulletCollision/CollisionDispatch/btSphereSphereCollisionAlgorithm.h"
#include "BulletCollision/NarrowPhaseCollision/btGjkEpa2.h"


#define	kQCPlugIn_Name				@"tLBulletSoftBodyPlugin"
#define	kQCPlugIn_Description		@"tLBulletSoftBodyPlugin"

#define kRenderFPS		60.0

#define WORLD_WIDTH		90
#define WORLD_HEIGHT	50
#define WORLD_DEPTH	    100
#define CUBE_HALF_EXTENTS 1.5
static float		width;
static float		height;
static float		aspect;

/// Bullet
#define TRIANGLE_SIZE 8.f
static btVector3*	gGroundVertices=0;
static int*	gGroundIndices=0;
static int gGroundTotalIndices = 0;
static float waveheight = 5.f;

const int maxProxies = 32766;

btSoftBodyWorldInfo	m_softBodyWorldInfo;

btAlignedObjectArray<btCollisionShape*>		m_collisionShapes;

btBroadphaseInterface*	m_broadphase;

btCollisionDispatcher*	m_dispatcher;

btConstraintSolver*	m_solver;

btDefaultCollisionConfiguration* m_collisionConfiguration;

btSoftRigidDynamicsWorld* m_dynamicsWorld;


@implementation tLBulletSoftBodyPlugIn

/*
Here you need to declare the input / output properties as dynamic as Quartz Composer will handle their implementation
@dynamic inputFoo, outputBar;
*/

@dynamic inputGravityX, inputGravityY, inputGravityZ;

+ (NSDictionary*) attributes
{
	/*
	Return a dictionary of attributes describing the plug-in (QCPlugInAttributeNameKey, QCPlugInAttributeDescriptionKey...).
	*/
	
	return [NSDictionary dictionaryWithObjectsAndKeys:kQCPlugIn_Name, QCPlugInAttributeNameKey, kQCPlugIn_Description, QCPlugInAttributeDescriptionKey, nil];
}

+ (NSDictionary*) attributesForPropertyPortWithKey:(NSString*)key
{
	/*
	Specify the optional attributes for property based ports (QCPortAttributeNameKey, QCPortAttributeDefaultValueKey...).
	*/
	
	return nil;
}

+ (QCPlugInExecutionMode) executionMode
{
	/*
	Return the execution mode of the plug-in: kQCPlugInExecutionModeProvider, kQCPlugInExecutionModeProcessor, or kQCPlugInExecutionModeConsumer.
	*/
	
	return kQCPlugInExecutionModeConsumer;
}

+ (QCPlugInTimeMode) timeMode
{
	/*
	Return the time dependency mode of the plug-in: kQCPlugInTimeModeNone, kQCPlugInTimeModeIdle or kQCPlugInTimeModeTimeBase.
	*/
	
	return kQCPlugInTimeModeTimeBase;
}

- (id) init
{
	if(self = [super init]) {
		/*
		Allocate any permanent resource required by the plug-in.
		*/
	}
	
	return self;
}

- (void) finalize
{
	/*
	Release any non garbage collected resources created in -init.
	*/
	
	[super finalize];
}

- (void) dealloc
{
	/*
	Release any resources created in -init.
	*/
	
	[super dealloc];
}

+ (NSArray*) plugInKeys
{
	/*
	Return a list of the KVC keys corresponding to the internal settings of the plug-in.
	*/
	
	return nil;
}

- (id) serializedValueForKey:(NSString*)key;
{
	/*
	Provide custom serialization for the plug-in internal settings that are not values complying to the <NSCoding> protocol.
	The return object must be nil or a PList compatible i.e. NSString, NSNumber, NSDate, NSData, NSArray or NSDictionary.
	*/
	
	return [super serializedValueForKey:key];
}

- (void) setSerializedValue:(id)serializedValue forKey:(NSString*)key
{
	/*
	Provide deserialization for the plug-in internal settings that were custom serialized in -serializedValueForKey.
	Deserialize the value, then call [self setValue:value forKey:key] to set the corresponding internal setting of the plug-in instance to that deserialized value.
	*/
	
	[super setSerializedValue:serializedValue forKey:key];
}

- (QCPlugInViewController*) createViewController
{
	/*
	Return a new QCPlugInViewController to edit the internal settings of this plug-in instance.
	You can return a subclass of QCPlugInViewController if necessary.
	*/
	
	return [[QCPlugInViewController alloc] initWithPlugIn:self viewNibName:@"Settings"];
}

@end

@implementation tLBulletSoftBodyPlugIn (Execution)

/// --- Physics methods ---

- (void) initPhysics
{
	///register some softbody collision algorithms on top of the default btDefaultCollisionConfiguration
	m_collisionConfiguration = new btSoftBodyRigidBodyCollisionConfiguration();

	m_dispatcher=0;
	m_dispatcher = new	btCollisionDispatcher(m_collisionConfiguration);
	m_softBodyWorldInfo.m_dispatcher = m_dispatcher;


	btVector3 worldAabbMin(-100,-100,-100);
	btVector3 worldAabbMax(100,100,100);

	m_broadphase = new btAxisSweep3(worldAabbMin,worldAabbMax,maxProxies);

	m_softBodyWorldInfo.m_broadphase = m_broadphase;

	btSequentialImpulseConstraintSolver* solver = new btSequentialImpulseConstraintSolver();

	m_solver = solver;

	m_dynamicsWorld = new btSoftRigidDynamicsWorld(m_dispatcher,m_broadphase,m_solver,m_collisionConfiguration);

	m_dynamicsWorld->getDispatchInfo().m_enableSPU = true;
	m_softBodyWorldInfo.m_sparsesdf.Initialize();

	m_dynamicsWorld->setGravity(btVector3(0,-10,0));
	m_softBodyWorldInfo.m_gravity.setValue(0,-10,0);
}


- (void) createBoulders
{
	for ( int i =0;i<10 ; i++)
	{
		const btVector3	pos=btVector3(-50 + i*10 ,25 ,0 );
		
		btSoftBody*	psb=btSoftBodyHelpers::CreateEllipsoid(m_softBodyWorldInfo,pos,btVector3(1,1,1)*8,512);
		psb->m_cfg.kDF=1.f;
		
		btSoftBody::Material*	pm=psb->appendMaterial();
		pm->m_kLST=	0.15f;	
		psb->generateBendingConstraints(3,pm);
		
		psb->m_cfg.piterations	=	2;
		psb->m_cfg.collisions	=	btSoftBody::fCollision::CL_SS + btSoftBody::fCollision::CL_RS;
		psb->randomizeConstraints();
		
		psb->generateClusters(512);	
		
		m_dynamicsWorld->addSoftBody(psb);
	}
}


- (void) createGround
{	
	btTransform transform;
	transform.setIdentity();
	
	//create concave ground mesh obj
	btCollisionShape* groundShape = 0;
	{
		int i;
		int j;
		
		const int NUM_VERTS_X = 30;
		const int NUM_VERTS_Y = 30;
		const int totalVerts = NUM_VERTS_X*NUM_VERTS_Y;
		const int totalTriangles = 2*(NUM_VERTS_X-1)*(NUM_VERTS_Y-1);
		
		gGroundVertices = new btVector3[totalVerts];
		gGroundTotalIndices = totalTriangles*3;
		gGroundIndices = new int[gGroundTotalIndices];
		
		btScalar offset(-50);
		
		for ( i=0;i<NUM_VERTS_X;i++)
		{
			for (j=0;j<NUM_VERTS_Y;j++)
			{
				gGroundVertices[i+j*NUM_VERTS_X].setValue((i-NUM_VERTS_X*0.5f)*TRIANGLE_SIZE,
														  //0.f,
														  waveheight*sinf((float)i)*cosf((float)j+offset),
														  (j-NUM_VERTS_Y*0.5f)*TRIANGLE_SIZE);
			}
		}
		
		int vertStride = sizeof(btVector3);
		int indexStride = 3*sizeof(int);
		
		int index=0;
		for ( i=0;i<NUM_VERTS_X-1;i++)
		{
			//NSLog(@"indices %i",sizeof(gGroundIndices));
			
			for (int j=0;j<NUM_VERTS_Y-1;j++)
			{
				gGroundIndices[index++] = j*NUM_VERTS_X+i;
				gGroundIndices[index++] = j*NUM_VERTS_X+i+1;
				gGroundIndices[index++] = (j+1)*NUM_VERTS_X+i+1;
				gGroundIndices[index++] = j*NUM_VERTS_X+i;
				gGroundIndices[index++] = (j+1)*NUM_VERTS_X+i+1;
				gGroundIndices[index++] = (j+1)*NUM_VERTS_X+i;
			}
		}
		
		btTriangleIndexVertexArray* indexVertexArrays = new btTriangleIndexVertexArray(totalTriangles,
																					   gGroundIndices,
																					   indexStride,
																					   totalVerts,(btScalar*) &gGroundVertices[0].x(),vertStride);
		
		bool useQuantizedAabbCompression = true;
		
		groundShape = new btBvhTriangleMeshShape(indexVertexArrays,useQuantizedAabbCompression);
		groundShape->setMargin(0.5);
		
		m_collisionShapes.push_back(groundShape);
		transform.setOrigin(btVector3(0, 0, 0));
		
		btCollisionObject* groundOb = new btCollisionObject();
		groundOb->setWorldTransform(transform);
		groundOb->setInterpolationWorldTransform( transform);
		
		groundOb->setCollisionShape(groundShape);
		m_dynamicsWorld->addCollisionObject(groundOb);
	}
}

- (void) createBoundingBox
{
	btTransform transform;
	transform.setIdentity();
	
	
	// Top wall
	{
		btCollisionShape* topBox = new btBoxShape (btVector3(WORLD_WIDTH,CUBE_HALF_EXTENTS,WORLD_DEPTH));
		m_collisionShapes.push_back(topBox);
		
		transform.setOrigin(btVector3(0,WORLD_HEIGHT,0));
		
		btCollisionObject* topOb = new btCollisionObject();
		topOb->setWorldTransform(transform);
		topOb->setInterpolationWorldTransform( transform);
		
		topOb->setCollisionShape(topBox);
		m_dynamicsWorld->addCollisionObject(topOb);
	}
	
	
	// Left wall
	{
		btCollisionShape* leftBox = new btBoxShape (btVector3(CUBE_HALF_EXTENTS,WORLD_HEIGHT,WORLD_DEPTH));
		m_collisionShapes.push_back(leftBox);
		
		transform.setOrigin(btVector3(-WORLD_WIDTH,0,0));
		
		btCollisionObject* leftOb = new btCollisionObject();
		leftOb->setWorldTransform(transform);
		leftOb->setInterpolationWorldTransform( transform);
		
		leftOb->setCollisionShape(leftBox);
		m_dynamicsWorld->addCollisionObject(leftOb);
	}
	
	// Right wall
	{
		btCollisionShape* rightBox = new btBoxShape (btVector3(CUBE_HALF_EXTENTS,WORLD_HEIGHT,WORLD_DEPTH));
		m_collisionShapes.push_back(rightBox);
		
		transform.setOrigin(btVector3(WORLD_WIDTH,0,0));
		
		btCollisionObject* rightOb = new btCollisionObject();
		rightOb->setWorldTransform(transform);
		rightOb->setInterpolationWorldTransform( transform);
		
		rightOb->setCollisionShape(rightBox);
		m_dynamicsWorld->addCollisionObject(rightOb);
	}
	
	// Back wall
	{
		btCollisionShape* backBox = new btBoxShape (btVector3(WORLD_WIDTH,WORLD_HEIGHT,CUBE_HALF_EXTENTS));
		m_collisionShapes.push_back(backBox);
		
		transform.setOrigin(btVector3(0,0,-WORLD_DEPTH));
		
		btCollisionObject* backOb = new btCollisionObject();
		backOb->setWorldTransform(transform);
		backOb->setInterpolationWorldTransform( transform);
		
		backOb->setCollisionShape(backBox);
		m_dynamicsWorld->addCollisionObject(backOb);
	}
	
	// Front wall
	{
		btCollisionShape* frontBox = new btBoxShape (btVector3(WORLD_WIDTH,WORLD_HEIGHT,CUBE_HALF_EXTENTS));
		m_collisionShapes.push_back(frontBox);
		
		transform.setOrigin(btVector3(0,0,WORLD_DEPTH));
		
		btCollisionObject* frontOb = new btCollisionObject();
		frontOb->setWorldTransform(transform);
		frontOb->setInterpolationWorldTransform( transform);
		
		frontOb->setCollisionShape(frontBox);
		m_dynamicsWorld->addCollisionObject(frontOb);
	}
}



- (void) exitPhysics
{
	//cleanup in the reverse order of creation/initialization
	
	for(int i=m_dynamicsWorld->getNumCollisionObjects()-1;i>=0;i--)
	{
		btCollisionObject*	obj=m_dynamicsWorld->getCollisionObjectArray()[i];
		btRigidBody*		body=btRigidBody::upcast(obj);
		if(body&&body->getMotionState())
		{
			delete body->getMotionState();
		}
		while(m_dynamicsWorld->getNumConstraints())
		{
			btTypedConstraint*	pc=m_dynamicsWorld->getConstraint(0);
			m_dynamicsWorld->removeConstraint(pc);
			delete pc;
		}
		btSoftBody* softBody = btSoftBody::upcast(obj);
		if (softBody)
		{
			m_dynamicsWorld->removeSoftBody(softBody);
		} else
		{
			btRigidBody* body = btRigidBody::upcast(obj);
			if (body)
				m_dynamicsWorld->removeRigidBody(body);
			else
				m_dynamicsWorld->removeCollisionObject(obj);
		}
		delete obj;
	}
	
	
	//delete dynamics world
	delete m_dynamicsWorld;
	
	//delete solver
	delete m_solver;
	
	//delete broadphase
	delete m_broadphase;
	
	//delete dispatcher
	delete m_dispatcher;
	
	
	
	delete m_collisionConfiguration;
	
}

/// --- Drawing Methods ---

- (void) initGLScene: (CGLContextObj) cgl_ctx
{
	
	// Enable a single OpenGL light.
	GLfloat mat_shininess[] = { 5.0 };
	GLfloat mat_specular[] = { 0.5, 0.5, 0.5, 1.0 };
	glShadeModel(GL_SMOOTH);
		
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, mat_shininess);
		
	glEnable(GL_COLOR_MATERIAL);
	glEnable(GL_LIGHT0);
		
	// Enable back-face culling 
	glCullFace(GL_BACK);
	glEnable(GL_CULL_FACE);
		
	// Enable Depth test
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,  GL_ONE_MINUS_SRC_ALPHA);
		
	// Enable Depth test
	glEnable(GL_DEPTH_TEST);
}

- (void) renderGLScene: (CGLContextObj) cgl_ctx
{
	// Lock the Gl context
	CGLLockContext(cgl_ctx);
	
	// save our current GL state
	glPushAttrib(GL_COLOR_BUFFER_BIT | GL_TRANSFORM_BIT | GL_VIEWPORT_BIT);
	
	//Set OpenGl state
	{		
		// Clear to black
		glClearColor(0.0,0.0,0.0,1.0);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		
		// Set light position 
		GLfloat light_position[] = { 0, 0 , WORLD_DEPTH*3 , 0.0 };
		glLightfv(GL_LIGHT0, GL_POSITION, light_position);
		
		//Save & configure projection matrix
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		
		gluPerspective(45, //view angle
					   aspect,	//aspect ratio
					   1.0, //near clip
					   WORLD_DEPTH*3.f);//far clip
		
		
		//Save & configure model view matrix
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
		gluLookAt(0.0, 10.0, 120,  /* eye is at (0,0,5) */
				  0.0, 0.0, 0.0,      /* center is at (0,0,0) */
				  0.0, 1.0, 0.);      /* up is in positive Y direction */
		
	}
	
	[self drawGroundBox:cgl_ctx];
	[self drawBoulders:cgl_ctx];
	
	
	// Restore OpenGL states
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	
	// Restore OpenGl Attributs
	glPopAttrib();
	
	// Lock the gl contaxt
	CGLUnlockContext(cgl_ctx);
	
}

- (void) drawGroundBox: (CGLContextObj) cgl_ctx
{
	btTransform tr;
	btCollisionObject* obj0 = m_dynamicsWorld->getCollisionObjectArray()[0];
	tr = obj0->getWorldTransform(); 
	btVector3 pos = tr.getOrigin();
	
	glEnable(GL_LIGHTING);
	glColor4f(1.f, 1.f, 1.f, 1.f);
	
	glPushMatrix();
	{
		// Translate
		glTranslatef(pos.getX(), pos.getY(), pos.getZ());
		
		for (int i=0;i<gGroundTotalIndices;i+=3)
		{
			const btVector3& wv0 = gGroundVertices[gGroundIndices[i]];
			const btVector3& wv1 = gGroundVertices[gGroundIndices[i+1]];
			const btVector3& wv2 = gGroundVertices[gGroundIndices[i+2]];
			
			
			btVector3 normal = (wv1-wv0).cross(wv2-wv0);
			normal.normalize ();
			glNormal3f(-normal.getX(),-normal.getY(),-normal.getZ());
			
			glBegin (GL_LINE_LOOP);
			{
				glVertex3f (wv0.x(), wv0.y(), wv0.z());
				glVertex3f (wv1.x(), wv1.y(), wv1.z());
				glVertex3f (wv2.x(), wv2.y(), wv2.z());
			}
			glEnd();
		}
	}
	glPopMatrix();
	
	glDisable(GL_LIGHTING);
}

- (void) drawBoulders: (CGLContextObj) cgl_ctx
{
	glEnable(GL_LIGHTING);
	
	btSoftRigidDynamicsWorld* softWorld = (btSoftRigidDynamicsWorld*)m_dynamicsWorld;
	
	for (int i=0;i<softWorld->getSoftBodyArray().size();i++)
	{
		btSoftBody*	psb=(btSoftBody*)softWorld->getSoftBodyArray()[i];
		
		const btScalar	scl=(btScalar)1.f;
		
		glColor4f(1.f, 1.f, 1.f, 1.f);
		for(int j=0;j<psb->m_faces.size();++j)
		{
			const btSoftBody::Face&	f=psb->m_faces[j];
			
			const btVector3			x[]={f.m_n[0]->m_x,f.m_n[1]->m_x,f.m_n[2]->m_x};
			const btVector3			c=(x[0]+x[1]+x[2])/3;
			
			const btVector3& wv0 = (x[0]-c)*scl+c;
			const btVector3& wv1 = (x[1]-c)*scl+c;
			const btVector3& wv2 = (x[2]-c)*scl+c;
			
			btVector3 normal = (wv1-wv0).cross(wv2-wv0);
			normal.normalize ();
			glNormal3f(normal.getX(),normal.getY(),normal.getZ());
			
			glColor4f(0.75f, 0.75f, 0.75f, 1.f);
			glBegin (GL_TRIANGLES);
			{
				glVertex3f (wv0.x(), wv0.y(), wv0.z());
				glVertex3f (wv1.x(), wv1.y(), wv1.z());
				glVertex3f (wv2.x(), wv2.y(), wv2.z());
			}
			glEnd();
			
			glColor4f(0.5f, 0.5f, 0.5f, 1.f);
			glBegin (GL_LINE_LOOP);
			{
				glVertex3f (wv0.x(), wv0.y(), wv0.z());
				glVertex3f (wv1.x(), wv1.y(), wv1.z());
				glVertex3f (wv2.x(), wv2.y(), wv2.z());
			}
			glEnd();
		}
		
	}
	glDisable(GL_LIGHTING);
	
}


/// --- Plugin Methods ---

- (BOOL) startExecution:(id<QCPlugInContext>)context
{
	/*
	Called by Quartz Composer when rendering of the composition starts: perform any required setup for the plug-in.
	Return NO in case of fatal failure (this will prevent rendering of the composition to start).
	*/
	
	return YES;
}

- (void) enableExecution:(id<QCPlugInContext>)context
{
	/*
	Called by Quartz Composer when the plug-in instance starts being used by Quartz Composer.
	*/
	
	CGLContextObj cgl_ctx = [context CGLContextObj];
	
	// Get Context Size and aspect
	width = context.bounds.size.width;
	height = context.bounds.size.height;
	
	aspect = width / height ;
	
	[self initGLScene: cgl_ctx ];
	[self initPhysics];
	[self createGround];
	[self createBoundingBox];
	[self createBoulders];
}

- (BOOL) execute:(id<QCPlugInContext>)context atTime:(NSTimeInterval)time withArguments:(NSDictionary*)arguments
{
	/*
	Called by Quartz Composer whenever the plug-in instance needs to execute.
	Only read from the plug-in inputs and produce a result (by writing to the plug-in outputs or rendering to the destination OpenGL context) within that method and nowhere else.
	Return NO in case of failure during the execution (this will prevent rendering of the current frame to complete).
	
	The OpenGL context for rendering can be accessed and defined for CGL macros using:
	CGLContextObj cgl_ctx = [context CGLContextObj];
	*/
	
	
	// --- Set the Gl context ---
	CGLContextObj cgl_ctx = [context CGLContextObj];
	
	// --- Updates ---
	
	// updates view dimensions
	width = context.bounds.size.width;
	height = context.bounds.size.height;
	aspect = width / height ;
	
	// update gravity
	m_dynamicsWorld->setGravity(btVector3(self.inputGravityX,self.inputGravityY,self.inputGravityZ));
	m_softBodyWorldInfo.m_gravity.setValue(self.inputGravityX,self.inputGravityY,self.inputGravityZ);
	
	// --- Step simulation 
	m_dynamicsWorld->stepSimulation(1/kRenderFPS,10);
	
	// --- Render GL scene ---
	[self renderGLScene:cgl_ctx];
	
	return YES;
}

- (void) disableExecution:(id<QCPlugInContext>)context
{
	/*
	Called by Quartz Composer when the plug-in instance stops being used by Quartz Composer.
	*/
	
	[self exitPhysics];
}

- (void) stopExecution:(id<QCPlugInContext>)context
{
	/*
	Called by Quartz Composer when rendering of the composition stops: perform any required cleanup for the plug-in.
	*/
}

@end
